package de.espirit.firstspirit.opt.example.browserstack;

import de.espirit.common.base.Logging;
import de.espirit.common.json.JsonObjects;
import de.espirit.common.json.JsonTokener;
import de.espirit.common.tools.Objects;
import de.espirit.common.tools.Streams;
import de.espirit.common.tools.XmlDocuments;
import de.espirit.firstspirit.io.FileHandle;
import de.espirit.firstspirit.io.FileSystem;
import de.espirit.firstspirit.module.Configuration;
import de.espirit.firstspirit.module.ProjectEnvironment;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;
import info.clearthought.layout.TableLayoutConstraints;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JViewport;
import javax.swing.ListSelectionModel;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.table.DefaultTableModel;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.KeyboardFocusManager;
import java.awt.Window;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOError;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * Provides configuration functionality for BrowserStack External Preview items.
 *
 * The configuration GUI can be accessed from the project application component in project settings, and this class
 * also handles loading of configuration for use by other classes.
 */
public class BrowserStackConfiguration implements Configuration<ProjectEnvironment> {

	private static final String CONFIG_FILE = "browserstack.xml";

	private BrowserStackSetup _setup;
	private ProjectEnvironment _env;
	private JPanel _component;


	@Override
	public boolean hasGui() {
		return true;
	}


	@Nullable
	@Override
	public JComponent getGui(final Frame masterFrame) {
		if (_component == null) {
			_component = new JPanel();
			_component.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

			final TableLayout layout = new TableLayout();
			layout.setHGap(25);
			layout.setVGap(5);
			layout.setColumn(new double[] { TableLayout.PREFERRED, TableLayout.FILL });
			_component.setLayout(layout);

			int row = -1;

			// Basic configuration of the BrowserStack integration

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("full_screen"), new TableLayoutConstraints(0, row));
			final JCheckBox full_screen = new JCheckBox("enable full screen", _setup.full_screen);
			full_screen.addItemListener(e -> _setup.full_screen = full_screen.isSelected());
			_component.add(full_screen, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("scale_to_fit"), new TableLayoutConstraints(0, row));
			final JCheckBox scale_to_fit = new JCheckBox("enable scale to fit mode", _setup.scale_to_fit);
			scale_to_fit.addItemListener(e -> _setup.scale_to_fit = scale_to_fit.isSelected());
			_component.add(scale_to_fit, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("start"), new TableLayoutConstraints(0, row));
			final JCheckBox start = new JCheckBox("enable auto start mode", _setup.start);
			start.addItemListener(e -> _setup.start = start.isSelected());
			_component.add(start, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("speed"), new TableLayoutConstraints(0, row));
			final JCheckBox speed = new JCheckBox("prefer fast speed over high quality", _setup.speed == 1);
			speed.addItemListener(e -> _setup.speed = speed.isSelected() ? 1 : 2);
			_component.add(speed, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("access_id"), new TableLayoutConstraints(0, row));
			final JTextField access_id = new JTextField(_setup.access_id);
			access_id.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.access_id = access_id.getText();
				}
			});
			_component.add(access_id, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("access_key"), new TableLayoutConstraints(0, row));
			final JTextField access_key = new JTextField(_setup.access_key);
			access_key.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.access_key = access_key.getText();
				}
			});
			_component.add(access_key, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("orientation"), new TableLayoutConstraints(0, row));
			final JTextField orientation = new JTextField(_setup.orientation);
			orientation.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.orientation = orientation.getText();
				}
			});
			_component.add(orientation, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("quality"), new TableLayoutConstraints(0, row));
			final JTextField quality = new JTextField(_setup.quality);
			quality.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.quality = quality.getText();
				}
			});
			_component.add(quality, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("win_res"), new TableLayoutConstraints(0, row));
			final JTextField win_res = new JTextField(_setup.win_res);
			win_res.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.win_res = win_res.getText();
				}
			});
			_component.add(win_res, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("mac_res"), new TableLayoutConstraints(0, row));
			final JTextField mac_res = new JTextField(_setup.mac_res);
			mac_res.getDocument().addDocumentListener(new DocumentModificationListener() {
				@Override
				public void onModification() {
					_setup.mac_res = mac_res.getText();
				}
			});
			_component.add(mac_res, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("wait_time"), new TableLayoutConstraints(0, row));
			final JComboBox wait_time = new JComboBox(new DefaultComboBoxModel(new Integer[] {2, 5, 10, 15, 20, 60}));
			wait_time.setSelectedItem(_setup.wait_time);
			wait_time.addItemListener(e -> {
				final Object selectedItem = wait_time.getSelectedItem();
				_setup.wait_time = selectedItem instanceof Integer ? (Integer) selectedItem : 2;
			});
			_component.add(wait_time, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			_component.add(new JLabel("use_screenshot_api"), new TableLayoutConstraints(0, row));
			final JCheckBox use_screenshot_api = new JCheckBox("enable use of screenshot api", _setup.use_screenshot_api);
			use_screenshot_api.addItemListener(e -> _setup.use_screenshot_api = use_screenshot_api.isSelected());
			_component.add(use_screenshot_api, new TableLayoutConstraints(1, row));

			// Insert a component for the configuration of the environments/devices

			layout.insertRow(++row, TableLayoutConstants.FILL);
			final EnvironmentTableModel model = new EnvironmentTableModel();
			final JTable environments = new JTable(model);
			environments.getTableHeader().setReorderingAllowed(false);
			environments.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			environments.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseClicked(final MouseEvent e) {
					if (e.getClickCount() == 2) {
						final int selectedRow = environments.getSelectedRow();
						final BrowserStackSetup.BrowserStackEnvironment environment = (BrowserStackSetup.BrowserStackEnvironment) environments.getValueAt(selectedRow, -1);
						if (environment != null) {
							final EnvironmentDialog dialog = new EnvironmentDialog(getActiveWindow(), environment);
							if (dialog.getResult()) {
								model.fireTableDataChanged();
							}
						}
					}
				}
			});

			final JScrollPane scrollPane = new JScrollPane(environments);
			scrollPane.setRowHeader(new JViewport());
			_component.add(scrollPane, new TableLayoutConstraints(0, row, 1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			final JPanel buttons = new JPanel(new FlowLayout(FlowLayout.CENTER, 5, 5));
			final JButton add = new JButton("Add");
			add.addActionListener(e -> {
				final BrowserStackSetup.BrowserStackEnvironment environment = new BrowserStackSetup.BrowserStackEnvironment();
				final EnvironmentDialog dialog = new EnvironmentDialog(getActiveWindow(), environment);
				if (dialog.getResult()) {
					_setup.environments.add(environment);
					model.fireTableDataChanged();
					final int index = _setup.environments.size() - 1;
					environments.setRowSelectionInterval(index, index);
				}
			});
			buttons.add(add);
			final JButton remove = new JButton("Remove");
			remove.setEnabled(false);
			remove.addActionListener(e -> {
				final int selectedRow = environments.getSelectedRow();
				final BrowserStackSetup.BrowserStackEnvironment environment = (BrowserStackSetup.BrowserStackEnvironment) environments.getValueAt(selectedRow, -1);
				if (environment != null) {
					_setup.environments.remove(environment);
					model.fireTableDataChanged();
				}
			});
			buttons.add(remove);
			final JButton edit = new JButton("Edit");
			edit.setEnabled(false);
			edit.addActionListener(e -> {
				final int selectedRow = environments.getSelectedRow();
				final BrowserStackSetup.BrowserStackEnvironment environment = (BrowserStackSetup.BrowserStackEnvironment) environments.getValueAt(selectedRow, -1);
				if (environment != null) {
					final EnvironmentDialog dialog = new EnvironmentDialog(getActiveWindow(), environment);
					if (dialog.getResult()) {
						model.fireTableDataChanged();
					}
				}
			});
			buttons.add(edit);
			_component.add(buttons, new TableLayoutConstraints(0, row, 1, row));

			environments.getSelectionModel().addListSelectionListener(e -> {
				final int selectedRow = environments.getSelectedRow();
				final BrowserStackSetup.BrowserStackEnvironment environment = (BrowserStackSetup.BrowserStackEnvironment) environments.getValueAt(selectedRow, -1);
				remove.setEnabled(environment != null);
				edit.setEnabled(environment != null);
			});
		}
		return _component;
	}


	@Nullable
	private static Window getActiveWindow() {
		Window window = KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow();
		while (window != null && !window.isVisible()) {
			window = window.getOwner();
		}
		return window;
	}


	@Override
	public void load() {
		_setup = loadConfiguration(_env.getConfDir());
	}


	@NotNull
	static BrowserStackSetup loadConfiguration(@NotNull final FileSystem<? extends FileHandle> confDir) {
		try {
			final FileHandle handle = confDir.obtain("/");
			final FileHandle child = handle.getChild(CONFIG_FILE);
			if (child.isFile()) {
				try (final InputStream in = child.load()) {
					try {
						final Document document = XmlDocuments.parse(in);
						return BrowserStackSetup.parseXML(document);
					} catch (final Exception e) {
						Logging.logWarning("Failed loading " + CONFIG_FILE, e, BrowserStackConfiguration.class);
					}
				}
			}
		} catch (final IOException e) {
			throw new IOError(e);
		}
		return new BrowserStackSetup();
	}


	@Override
	public void store() {
		try {
			final FileHandle handle = _env.getConfDir().obtain("/");
			final FileHandle contextHandle = handle.getChild(CONFIG_FILE);
			try (final OutputStream out = contextHandle.getOutputStream()) {
				final String xml = _setup.getXML();
				out.write(xml.getBytes(StandardCharsets.UTF_8));
			} catch (final Exception e) {
				Logging.logWarning("Failed storing " + CONFIG_FILE, e, BrowserStackConfiguration.class);
			}
		} catch (final IOException e) {
			throw new IOError(e);
		}
	}


	@Override
	public Set<String> getParameterNames() {
		return Collections.emptySet();
	}


	@Nullable
	@Override
	public String getParameter(final String name) {
		return null;
	}


	@Override
	public void init(final String moduleName, final String componentName, final ProjectEnvironment env) {
		_env = env;
	}


	@Override
	public ProjectEnvironment getEnvironment() {
		return _env;
	}


	/**
	 * Table model for the configurable list of environments to be accessible by the user.
	 */
	@SuppressWarnings("SerializableInnerClassWithNonSerializableOuterClass")
	private class EnvironmentTableModel extends DefaultTableModel {

		private static final long serialVersionUID = 1L;


		EnvironmentTableModel() {
			addColumn("os");
			addColumn("os_version");
			addColumn("browser");
			addColumn("browser_version");
			addColumn("device");
		}


		@Override
		public int getRowCount() {
			return _setup.environments.size();
		}


		@Nullable
		@Override
		public Object getValueAt(final int row, final int column) {
			if (row > -1 && row < _setup.environments.size()) {
				final BrowserStackSetup.BrowserStackEnvironment environment = _setup.environments.get(row);
				if (column == -1) {
					return environment;
				} else if (column == 0) {
					return environment.os;
				} else if (column == 1) {
					return environment.os_version;
				} else if (column == 2) {
					return environment.browser;
				} else if (column == 3) {
					return environment.browser_version;
				} else if (column == 4) {
					return environment.device;
				}
			}
			return null;
		}


		@Override
		public boolean isCellEditable(final int row, final int column) {
			return false;
		}
	}


	/**
	 * Detail dialog for the configuration of a single environment/device from BrowserStack.
	 */
	private class EnvironmentDialog extends JDialog {

		private boolean _modified;


		EnvironmentDialog(@Nullable final Window owner, final BrowserStackSetup.BrowserStackEnvironment environment) {
			super(owner);
			setModal(true);

			final JPanel component = new JPanel();
			component.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

			final TableLayout layout = new TableLayout();
			layout.setHGap(25);
			layout.setVGap(5);
			layout.setColumn(new double[] { TableLayout.PREFERRED, TableLayout.FILL });
			component.setLayout(layout);

			int row = -1;

			final JTextField os = new JTextField(environment.os);
			final JTextField os_version = new JTextField(environment.os_version);
			final JTextField browser = new JTextField(environment.browser);
			final JTextField browser_version = new JTextField(environment.browser_version);
			final JTextField device = new JTextField(environment.device);

			// Insert a preset list using the BrowserStack API
			final List<BrowserStackSetup.BrowserStackEnvironment> environments = getEnvironments();
			if (environments != null && !environments.isEmpty()) {
				layout.insertRow(++row, TableLayoutConstants.PREFERRED);
				component.add(new JLabel("preset"), new TableLayoutConstraints(0, row));
				final List<Object> presets = new ArrayList<>();
				presets.add("");
				presets.addAll(environments);
				final DefaultComboBoxModel model = new DefaultComboBoxModel(presets.toArray(new Object[presets.size()]));
				final JComboBox preset = new JComboBox(model);
				preset.addItemListener(e -> {
					final Object selection = preset.getSelectedItem();
					if (selection instanceof BrowserStackSetup.BrowserStackEnvironment) {
						final BrowserStackSetup.BrowserStackEnvironment use = (BrowserStackSetup.BrowserStackEnvironment) selection;
						os.setText(Objects.nvl(use.os, ""));
						os_version.setText(Objects.nvl(use.os_version, ""));
						browser.setText(Objects.nvl(use.browser, ""));
						browser_version.setText(Objects.nvl(use.browser_version, ""));
						device.setText(Objects.nvl(use.device, ""));
					}
				});
				component.add(preset, new TableLayoutConstraints(1, row));
			}

			// Insert the fields for the environment attributes

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			component.add(new JLabel("os"), new TableLayoutConstraints(0, row));
			component.add(os, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			component.add(new JLabel("os_version"), new TableLayoutConstraints(0, row));
			component.add(os_version, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			component.add(new JLabel("browser"), new TableLayoutConstraints(0, row));
			component.add(browser, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			component.add(new JLabel("browser_version"), new TableLayoutConstraints(0, row));
			component.add(browser_version, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			component.add(new JLabel("device"), new TableLayoutConstraints(0, row));
			component.add(device, new TableLayoutConstraints(1, row));

			layout.insertRow(++row, TableLayoutConstants.FILL);
			component.add(Box.createVerticalGlue(), new TableLayoutConstraints(0, row));

			// Ok & Cancel
			layout.insertRow(++row, TableLayoutConstants.PREFERRED);
			final JPanel buttons = new JPanel(new FlowLayout(FlowLayout.CENTER, 10, 0));
			final JButton ok = new JButton("Ok");
			ok.addActionListener(e -> {
				_modified = true;
				environment.os = os.getText();
				environment.os_version = os_version.getText();
				environment.browser = browser.getText();
				environment.browser_version = browser_version.getText();
				environment.device = device.getText();
				setVisible(false);
			});
			buttons.add(ok);
			final JButton cancel = new JButton("Cancel");
			cancel.addActionListener(e -> setVisible(false));
			buttons.add(cancel);
			component.add(buttons, new TableLayoutConstraints(0, row, 1, row));

			setContentPane(component);
		}


		/**
		 * Opens this modal detail dialog.
		 *
		 * @return Whether or not the configured device has been modified.
		 */
		boolean getResult() {
			pack();
			setLocationRelativeTo(getOwner());
			setVisible(true);
			return _modified;
		}


		/**
		 * Requests a list of available environments/devices from BrowserStack using its API.
		 *
		 * @return A list of available devices.
		 */
		@Nullable
		List<BrowserStackSetup.BrowserStackEnvironment> getEnvironments() {
			try {
				// Request devices as JSON
				final URL url = new URL("http://" + _setup.access_id + ':' + _setup.access_key + "@www.browserstack.com/screenshots/browsers.json");
				final URLConnection connection = url.openConnection();

				// Parse JSON response
				final InputStream stream = connection.getInputStream();
				final String string = Streams.toString(stream, "UTF-8");
				final JsonTokener tokener = new JsonTokener(string);
				final Object json = tokener.nextValue();
				final Object object = JsonObjects.fromJson(json);
				if (object instanceof Iterable<?>) {

					// Create our internal objects based on the device list from BrowserStack
					final List<BrowserStackSetup.BrowserStackEnvironment> result = new ArrayList<>();
					for (final Object map : (Iterable<?>) object) {
						if (map instanceof Map<?, ?>) {
							final BrowserStackSetup.BrowserStackEnvironment env = new BrowserStackSetup.BrowserStackEnvironment();
							env.os = (String) ((Map<?, ?>) map).get("os");
							env.os_version = (String) ((Map<?, ?>) map).get("os_version");
							env.browser = (String) ((Map<?, ?>) map).get("browser");
							env.browser_version = (String) ((Map<?, ?>) map).get("browser_version");
							env.device = (String) ((Map<?, ?>) map).get("device");
							result.add(env);
						}
					}
					return result;
				}
			} catch (final Throwable e) {
				Logging.logWarning(e.getMessage(), e, ScreenshotsItem.class);
			}
			return null;
		}
	}


	/**
	 * Listener for document modification events.
	 */
	private abstract static class DocumentModificationListener implements DocumentListener {

		@Override
		public void insertUpdate(final DocumentEvent e) {
			onModification();
		}


		@Override
		public void removeUpdate(final DocumentEvent e) {
			onModification();
		}


		@Override
		public void changedUpdate(final DocumentEvent e) {
			onModification();
		}


		public abstract void onModification();
	}
}
