package de.espirit.firstspirit.opt.example.browserstack;

import de.espirit.firstspirit.access.BaseContext;
import de.espirit.firstspirit.access.project.Project;
import de.espirit.firstspirit.access.store.Store;
import de.espirit.firstspirit.agency.ModuleAdminAgent;
import de.espirit.firstspirit.agency.StoreAgent;
import de.espirit.firstspirit.client.plugin.ExternalPreviewItem;
import de.espirit.firstspirit.client.plugin.ExternalPreviewItemsPlugin;
import de.espirit.firstspirit.io.FileSystem;

import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;


/**
 * This class configures and provides {@link ExternalPreviewItem} objects which may be shown in the dropdown list associated
 * with the Preview button in SiteArchitect's editorial toolbar.
 */
public class BrowserStackItemsPlugin implements ExternalPreviewItemsPlugin {

	private BrowserStackSetup _setup;


	/**
	 * Provide available {@link ExternalPreviewItem} objects iff the project application has been added to the current project
	 * and a non-null {@link BrowserStackSetup} object can be obtained (see {@link #setUp(BaseContext)}).
	 *
	 * @return A {@link Collection} of {@link ExternalPreviewItem}s.
	 */
	@NotNull
	@Override
	public Collection<? extends ExternalPreviewItem> getItems() {
		final List<ExternalPreviewItem> items = new ArrayList<ExternalPreviewItem>();

		// External Preview items for BrowserStack use require a BrowserStackSetup object (field _setup).
		if (_setup != null) {
			for (final BrowserStackSetup.BrowserStackEnvironment env : _setup.environments) {
				items.add(new LiveItem(_setup, env));
			}
			if (_setup.environments.size() > 1) {
				items.add(new ScreenshotsItem(_setup));
			}
		}
		return items;
	}


	/**
	 * Setup for this class.
	 *
	 * Load the project application configuration, if possible.
	 *
	 * @param context The context the plug-in lives in.
	 */
	@Override
	public void setUp(@NotNull final BaseContext context) {
		final ModuleAdminAgent moduleAdminAgent = context.requireSpecialist(ModuleAdminAgent.TYPE);
		final Project project = context.requireSpecialist(StoreAgent.TYPE).getStore(Store.Type.PAGESTORE, false).getProject();
		try {
			final FileSystem<?> confDir = moduleAdminAgent.getProjectAppConfig("BrowserStack", "BrowserStackProjectConfiguration", project);
			_setup = BrowserStackConfiguration.loadConfiguration(confDir);
		} catch (final IllegalArgumentException ignored) {
		}
	}


	@Override
	public void tearDown() {
	}
}
