package de.espirit.firstspirit.opt.example.browserstack;

import de.espirit.common.base.Logging;
import de.espirit.common.tools.XmlDocuments;

import org.jetbrains.annotations.NotNull;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.util.ArrayList;
import java.util.List;


/**
 * BrowserStack-specific functionality.
 *
 * See http://www.browserstack.com/integrations for more information.
 */
public class BrowserStackSetup {

	// http://www.browserstack.com/integrations
	// http://www.browserstack.com/screenshots/api

	public final List<BrowserStackEnvironment> environments = new ArrayList<BrowserStackEnvironment>();
	public boolean full_screen;
	public boolean scale_to_fit = true;
	public boolean start;
	public boolean use_screenshot_api;
	public int speed = 1;

	public String access_id = "";
	public String access_key = "";
	public String orientation = "portrait";
	public String quality = "compressed";
	public String win_res = "1024x768";
	public String mac_res = "1024x768";
	public int wait_time = 2;


	/**
	 * Get the current setup configuration as XML.
	 *
	 * @return configuration as XML
	 * @see #parseXML(Document)
	 */
	@NotNull
	public String getXML() {
		final StringBuilder xml = new StringBuilder();
		xml.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		xml.append("<browserstack>\n");

		xml.append("  <environments>\n");
		for (final BrowserStackEnvironment environment : environments) {
			xml.append("    <environment>\n");
			xml.append("      <os>").append(XmlDocuments.normalizeText(environment.os)).append("</os>\n");
			xml.append("      <os_version>").append(XmlDocuments.normalizeText(environment.os_version)).append("</os_version>\n");
			xml.append("      <browser>").append(XmlDocuments.normalizeText(environment.browser)).append("</browser>\n");
			xml.append("      <browser_version>").append(XmlDocuments.normalizeText(environment.browser_version)).append("</browser_version>\n");
			xml.append("      <device>").append(XmlDocuments.normalizeText(environment.device)).append("</device>\n");
			xml.append("    </environment>");
		}
		xml.append("  </environments>\n");
		xml.append("  <full_screen>").append(full_screen).append("</full_screen>\n");
		xml.append("  <scale_to_fit>").append(scale_to_fit).append("</scale_to_fit>\n");
		xml.append("  <speed>").append(speed).append("</speed>\n");
		xml.append("  <start>").append(start).append("</start>\n");
		xml.append("  <access_id>").append(XmlDocuments.normalizeText(access_id)).append("</access_id>\n");
		xml.append("  <access_key>").append(XmlDocuments.normalizeText(access_key)).append("</access_key>\n");
		xml.append("  <orientation>").append(XmlDocuments.normalizeText(orientation)).append("</orientation>\n");
		xml.append("  <quality>").append(XmlDocuments.normalizeText(quality)).append("</quality>\n");
		xml.append("  <win_res>").append(XmlDocuments.normalizeText(win_res)).append("</win_res>\n");
		xml.append("  <mac_res>").append(XmlDocuments.normalizeText(mac_res)).append("</mac_res>\n");
		xml.append("  <wait_time>").append(wait_time).append("</wait_time>\n");
		xml.append("  <use_screenshot_api>").append(use_screenshot_api).append("</use_screenshot_api>\n");

		xml.append("</browserstack>");
		return xml.toString();
	}


	/**
	 * Parses the specified XML document and creates a {@link BrowserStackSetup}
	 * instance based on the containing configuration settings.
	 *
	 * @param document The XML document to parse.
	 * @return The related {@link BrowserStackSetup} instance.
	 * @see #getXML()
	 */
	@NotNull
	public static BrowserStackSetup parseXML(@NotNull final Document document) {
		final BrowserStackSetup setup = new BrowserStackSetup();
		try {
			final XPathFactory factory = XPathFactory.newInstance();

			final NodeList envList = (NodeList) factory.newXPath().evaluate("/browserstack/environments/environment", document, XPathConstants.NODESET);
			if (envList != null) {
				for (int i = 0; i < envList.getLength(); i++) {
					final Node env = envList.item(i);
					final BrowserStackEnvironment environment = new BrowserStackEnvironment();
					environment.os = nvl((String) factory.newXPath().evaluate("os/text()", env, XPathConstants.STRING), environment.os);
					environment.os_version = nvl((String) factory.newXPath().evaluate("os_version/text()", env, XPathConstants.STRING), environment.os_version);
					environment.browser = nvl((String) factory.newXPath().evaluate("browser/text()", env, XPathConstants.STRING), environment.browser);
					environment.browser_version = nvl((String) factory.newXPath().evaluate("browser_version/text()", env, XPathConstants.STRING), environment.browser_version);
					environment.device = nvl((String) factory.newXPath().evaluate("device/text()", env, XPathConstants.STRING), environment.device);
					setup.environments.add(environment);
				}
			}
			setup.full_screen = Boolean.parseBoolean((String) factory.newXPath().evaluate("/browserstack/full_screen/text()", document, XPathConstants.STRING));
			setup.scale_to_fit = Boolean.parseBoolean((String) factory.newXPath().evaluate("/browserstack/scale_to_fit/text()", document, XPathConstants.STRING));
			setup.start = Boolean.parseBoolean((String) factory.newXPath().evaluate("/browserstack/start/text()", document, XPathConstants.STRING));
			try {
				setup.speed = Integer.parseInt((String) factory.newXPath().evaluate("/browserstack/speed/text()", document, XPathConstants.STRING));
			} catch (final NumberFormatException ignored) {
			}
			setup.access_id = nvl((String) factory.newXPath().evaluate("/browserstack/access_id/text()", document, XPathConstants.STRING), setup.access_id);
			setup.access_key = nvl((String) factory.newXPath().evaluate("/browserstack/access_key/text()", document, XPathConstants.STRING), setup.access_key);
			setup.orientation = nvl((String) factory.newXPath().evaluate("/browserstack/orientation/text()", document, XPathConstants.STRING), setup.orientation);
			setup.quality = nvl((String) factory.newXPath().evaluate("/browserstack/quality/text()", document, XPathConstants.STRING), setup.quality);
			setup.win_res = nvl((String) factory.newXPath().evaluate("/browserstack/win_res/text()", document, XPathConstants.STRING), setup.win_res);
			setup.mac_res = nvl((String) factory.newXPath().evaluate("/browserstack/mac_res/text()", document, XPathConstants.STRING), setup.mac_res);
			try {
				setup.wait_time = Integer.parseInt((String) factory.newXPath().evaluate("/browserstack/wait_time/text()", document, XPathConstants.STRING));
			} catch (final NumberFormatException ignored) {
			}
			setup.use_screenshot_api = Boolean.parseBoolean((String) factory.newXPath().evaluate("/browserstack/use_screenshot_api/text()", document, XPathConstants.STRING));

		} catch (final XPathExpressionException e) {
			Logging.logWarning("Failed parsing BrowserStack configuration", BrowserStackSetup.class);
		}
		return setup;
	}


	private static String nvl(final String value, final String fallback) {
		return value != null && !value.isEmpty() ? value : fallback;
	}


	/**
	 * One environment/device in BrowserStack
	 */
	public static class BrowserStackEnvironment {

		public String os = "Windows";
		public String os_version = "XP";
		public String browser = "Firefox";
		public String browser_version = "10.0";
		public String device = "";


		@Override
		public String toString() {
			final StringBuilder string = new StringBuilder();
			if (device != null && !device.isEmpty()) {
				string.append(device);
			} else {
				string.append(os);
				if (os_version != null && !os_version.isEmpty()) {
					string.append(' ').append(os_version);
				}
				if (browser != null && !browser.isEmpty()) {
					string.append(": ");
					string.append(browser);
					if (browser_version != null && !browser_version.isEmpty()) {
						string.append(' ').append(browser_version);
					}
				}
			}
			return string.toString();
		}
	}
}
