package de.espirit.firstspirit.opt.example.browserstack;

import de.espirit.common.tools.Strings;
import de.espirit.firstspirit.access.Language;
import de.espirit.firstspirit.access.store.IDProvider;
import de.espirit.firstspirit.access.store.sitestore.PageRef;
import de.espirit.firstspirit.agency.GenerateElementOperation;
import de.espirit.firstspirit.agency.OperationAgent;
import de.espirit.firstspirit.client.gui.applications.browser.EngineType;
import de.espirit.firstspirit.client.plugin.ExternalPreviewContext;
import de.espirit.firstspirit.client.plugin.ExternalPreviewItem;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.Icon;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;


/**
 * Item which starts a LIVE debugging session for a specific environment/device in BrowserStack.
 */
public class LiveItem implements ExternalPreviewItem {

	private final BrowserStackSetup _setup;
	private final BrowserStackSetup.BrowserStackEnvironment _environment;


	public LiveItem(@NotNull final BrowserStackSetup setup, @NotNull final BrowserStackSetup.BrowserStackEnvironment environment) {
		_setup = setup;
		_environment = environment;
	}


	@Override
	public boolean isVisible(@NotNull final ExternalPreviewContext context) {
		return context.getElement() instanceof PageRef;
	}


	@Override
	public boolean isEnabled(@NotNull final ExternalPreviewContext context) {
		return context.getElement() instanceof PageRef;
	}


	@Override
	public String getLabel(@NotNull final ExternalPreviewContext context) {
		return "BrowserStack: " + _environment.toString();
	}


	@Nullable
	@Override
	public Icon getIcon(@NotNull final ExternalPreviewContext context) {
		return null;
	}


	@Nullable
	@Override
	public EngineType getBrowserType() {
		// BrowserStack only works in CHROME browser.
		return EngineType.CHROME;
	}


	@Override
	public String getUrl(@NotNull final ExternalPreviewContext context) {
		final IDProvider element = context.getElement();
		if (element instanceof PageRef) {
			Language language = context.getLanguage();
			if (language == null) {
				language = element.getProject().getMasterLanguage();
			}
			// Generate the current PageRef element, so we can access it from BrowserStack
			final GenerateElementOperation operation = context.requireSpecialist(OperationAgent.TYPE).getOperation(GenerateElementOperation.TYPE);
			if (operation != null) {
				final String targetUrl = operation.perform((PageRef) element, language);
				if (targetUrl != null) {
					// Create the BrowserStack URL using the target URL, based on the current BrowserStackSetup
					return createLiveUrl(targetUrl);
				}
			}
		}
		return null;
	}


	/**
	 * Creates a BrowserStack URL for a LIVE debugging session
	 * using the configured {@link BrowserStackSetup}.
	 *
	 * @param targetUrl The target URL to display.
	 * @return The BrowserStack URL.
	 */
	@NotNull
	private String createLiveUrl(@NotNull final String targetUrl) {
		final Map<String, String> parameter = new HashMap<String, String>();
		if (!Strings.isEmpty(_environment.os)) {
			parameter.put("os", _environment.os);
		}
		if (!Strings.isEmpty(_environment.os_version)) {
			parameter.put("os_version", _environment.os_version);
		}
		if (!Strings.isEmpty(_environment.browser)) {
			parameter.put("browser", _environment.browser);
		}
		if (!Strings.isEmpty(_environment.browser_version)) {
			parameter.put("browser_version", _environment.browser_version);
		}
		if (!Strings.isEmpty(_environment.device)) {
			parameter.put("device", _environment.device);
		}
		if ("windows".equalsIgnoreCase(_environment.os)) {
			parameter.put("resolution", _setup.win_res);
		} else if ("os x".equalsIgnoreCase(_environment.os)) {
			parameter.put("resolution", _setup.mac_res);
		}
		parameter.put("scale_to_fit", String.valueOf(_setup.scale_to_fit));
		parameter.put("full_screen", String.valueOf(_setup.full_screen));
		parameter.put("speed", String.valueOf(_setup.speed));
		parameter.put("start", String.valueOf(!_setup.start));
		parameter.put("url", targetUrl);

		try {
			final StringBuilder serviceUrl = new StringBuilder();
			serviceUrl.append("https://www.browserstack.com/start#");
			final Iterator<Map.Entry<String,String>> iterator = parameter.entrySet().iterator();
			while (iterator.hasNext()) {
				final Map.Entry<String, String> entry = iterator.next();
				serviceUrl.append(URLEncoder.encode(entry.getKey(), "UTF-8"));
				serviceUrl.append('=');
				serviceUrl.append(URLEncoder.encode(entry.getValue(), "UTF-8"));
				if (iterator.hasNext()) {
					serviceUrl.append('&');
				}
			}
			return serviceUrl.toString();
		} catch (final UnsupportedEncodingException e) {
			throw new IllegalStateException(e);
		}
	}
}
