package de.espirit.firstspirit.opt.example.browserstack;

import de.espirit.common.base.Logging;
import de.espirit.common.json.JsonObjects;
import de.espirit.common.json.JsonTokener;
import de.espirit.common.tools.Streams;
import de.espirit.firstspirit.access.Language;
import de.espirit.firstspirit.access.ServicesBroker;
import de.espirit.firstspirit.access.store.IDProvider;
import de.espirit.firstspirit.access.store.sitestore.PageRef;
import de.espirit.firstspirit.agency.GenerateElementOperation;
import de.espirit.firstspirit.agency.OperationAgent;
import de.espirit.firstspirit.client.gui.applications.ApplicationService;
import de.espirit.firstspirit.client.gui.applications.ApplicationTab;
import de.espirit.firstspirit.client.gui.applications.browser.BrowserApplication;
import de.espirit.firstspirit.client.gui.applications.browser.BrowserApplicationConfiguration;
import de.espirit.firstspirit.client.gui.applications.browser.BrowserListener;
import de.espirit.firstspirit.client.gui.applications.browser.EngineType;
import de.espirit.firstspirit.client.plugin.ExternalPreviewContext;
import de.espirit.firstspirit.client.plugin.ExternalPreviewItem;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.swing.Icon;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Item which uses the BrowserStack Screenshot API to produce a screenshot of the contextual PageRef element.
 */
public class ScreenshotsItem implements ExternalPreviewItem {

	private final BrowserStackSetup _setup;


	public ScreenshotsItem(final BrowserStackSetup setup) {
		_setup = setup;
	}


	@Override
	public String getLabel(@NotNull final ExternalPreviewContext context) {
		return "BrowserStack: Screenshots";
	}


	@Nullable
	@Override
	public Icon getIcon(@NotNull final ExternalPreviewContext context) {
		return null;
	}


	@Nullable
	@Override
	public EngineType getBrowserType() {
		// BrowserStack only works in CHROME browser.
		return EngineType.CHROME;
	}


	@Nullable
	@Override
	public String getUrl(@NotNull final ExternalPreviewContext context) {
		final IDProvider element = context.getElement();
		if (element instanceof PageRef) {
			Language language = context.getLanguage();
			if (language == null) {
				language = element.getProject().getMasterLanguage();
			}
			// Generate the current PageRef element, so we can access it from BrowserStack
			final GenerateElementOperation operation = context.requireSpecialist(OperationAgent.TYPE).getOperation(GenerateElementOperation.TYPE);
			if (operation != null) {
				final String targetUrl = operation.perform((PageRef) element, language);
				if (targetUrl != null) {

					if (_setup.use_screenshot_api && !_setup.environments.isEmpty()) {
						// Screenshot API is enabled in the configuration => use it
						return useScreenshotAPI(targetUrl);
					}

					// Unfortunately, we have no API access to the Screenshot API from BrowserStack
					// => Open a BrowserApplication and insert the target URL into the related URL field
					//    Note: We cannot preselect the configured devices automatically this way :(
					final ServicesBroker services = context.requireSpecialist(ServicesBroker.TYPE);
					final ApplicationService service = services.getService(ApplicationService.class);
					final BrowserApplicationConfiguration config = BrowserApplicationConfiguration.GENERATOR.invoke();
					final ApplicationTab<BrowserApplication> tab = service.openApplication(BrowserApplication.TYPE, config);
					final BrowserApplication application = tab.getApplication();
					application.addBrowserListener(new BrowserListener.Adapter() {
						@Override
						public void onDocumentComplete(final String url) {
							final Document document = application.getCurrentDocument();
							final Element field = document.getElementById("screenshots");
							if (field != null) {
								field.setAttribute("value", targetUrl);
							}
						}
					});
					application.openUrl("http://www.browserstack.com/screenshots");
				}
			}
		}
		return null;
	}


	/**
	 * Starts a job using the BrowserStack Screenshot API.
	 * This features needs to licensed and may not be available for the configured account.
	 *
	 * @param targetUrl The target URL to be rendered in different environments/devices.
	 * @return The URL to display in the integrated browser.
	 */
	@Nullable
	private String useScreenshotAPI(@NotNull final String targetUrl) {

		// configure Screenshot API request parameter
		final Map<String, Object> request = new HashMap<String, Object>();
		request.put("url", targetUrl);
		request.put("quality", _setup.quality);
		request.put("win_res", _setup.win_res);
		request.put("mac_res", _setup.mac_res);
		request.put("wait_time", _setup.wait_time);
		request.put("orientation", _setup.orientation);

		// configure devices
		final List<Map<String, Object>> browsers = new ArrayList<Map<String, Object>>();
		for (final BrowserStackSetup.BrowserStackEnvironment environment : _setup.environments) {
			final Map<String, Object> browser = new HashMap<String, Object>();
			browser.put("os", environment.os);
			browser.put("os_version", environment.os_version);
			browser.put("browser", environment.browser);
			browser.put("browser_version", environment.browser_version);
			browser.put("device", environment.device);
			browsers.add(browser);
		}
		request.put("browsers", browsers);

		try {
			// start a job using the BrowserStack Screenshot API
			final String screenshotUrl = "http://www.browserstack.com/screenshots";
			final URL url = new URL(screenshotUrl);
			final URLConnection connection = url.openConnection();
			connection.setDoInput(true);
			connection.setDoOutput(true);
			connection.setRequestProperty("Authorization", Streams.toBase64((_setup.access_id + ':' + _setup.access_key).getBytes("UTF-8")));
			connection.setRequestProperty("Content-Type", "application/json");
			connection.setRequestProperty("Accept", "application/json");
			final OutputStream outputStream = connection.getOutputStream();
			try {
				final Object json = JsonObjects.toJson(request);
				final String string = json.toString();
				outputStream.write(string.getBytes("UTF-8"));
			} finally {
				Streams.close(outputStream);
			}

			// parse the response from BrowserStack to get the job ID from
			final InputStream inputStream = connection.getInputStream();
			try {
				final String response = Streams.toString(inputStream, "UTF-8");
				final JsonTokener tokener = new JsonTokener(response);
				final Object json = tokener.nextValue();
				final Object object = JsonObjects.fromJson(json);
				if (object instanceof Map<?, ?>) {
					// return the URL for the screenshot results
					return screenshotUrl + '/' + ((Map<?, ?>) object).get("id");
				}
			} finally {
				Streams.close(inputStream);
			}
		} catch (final Throwable e) {
			Logging.logWarning(e.getMessage(), e, ScreenshotsItem.class);
		}
		return null;
	}


	@Override
	public boolean isVisible(@NotNull final ExternalPreviewContext context) {
		final IDProvider element = context.getElement();
		return element instanceof PageRef && _setup != null && !_setup.environments.isEmpty();
	}


	@Override
	public boolean isEnabled(@NotNull final ExternalPreviewContext context) {
		return true;
	}
}
